/***************************************************************************
Copyright 1999 The Regents of the University of California
All Rights Reserved

Permission to use, copy, modify and distribute any part of the
GTrace tool for educational, research and non-profit
purposes, without fee, and without a written agreement is hereby
granted, provided that the above copyright notice, this paragraph
and the following paragraphs appear in all copies.

Those desiring to incorporate this into commercial products or use
for commercial purposes should contact the Technology Transfer
Office, University of California, San Diego, 9500 Gilman Drive, La
Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.

THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
ANY PATENT, TRADEMARK OR OTHER RIGHTS.

The GTrace tool is developed by Ram Periakaruppan at the University 
of California, San Diego under the Cooperative Association for Internet 
Data Analysis (CAIDA) Program. Support for this effort is provided by 
the CAIDA grant ANI-9711092, and by CAIDA members.

Report bugs and suggestions to gtrace-bugs@caida.org.
**************************************************************************/

/******************************************************************************
**
**     File:                    DbTxt.java
**
**     Purpose:                 Used to store the details and contents of
**                              databases and text files.
**                              
**     Author/Date:             Ram Periakaruppan, Sept 15th, 1999.
**
**     References:
**      Naughton, P., Schildt, H., JAVA - The Complete Reference. McGraw Hill.
**      Eckstein, R., Loy, M., Wood, D., JAVA Swing. O'Reilly.
**      The Source For Java Technology, http://www.java.sun.com.
**
******************************************************************************/
package caida.GTrace;

import com.sleepycat.db.*;

public class DbTxt extends DbEnv
{
    private String DbTxtName;   // the database textfile name
    private int contents;       // its contents
    private int type;           // its type
    private GTraceDb gtraceDb;  // a GTraceDb instance to do db queries

    // Constants
    public static final int IPADDR = 1;
    public static final int HOSTNAME = 2;
    public static final int HOSTNAMEIPADDR = 3;

    public static final int DATABASE = 1;
    public static final int TEXTFILE = 2;
    
//---------------------------------------------------------------------------
//                             constructors
//---------------------------------------------------------------------------

/******************************************************************************
**
**     Purpose:                 Initializes the member variables to the
**                              parameters specified.
**    
**     Parameters:              String DbDir - database path
**                              String name - database name
**
**     Author/Last Modified:    Ram Periakaruppan, Sept 15th, 1999
**
******************************************************************************/

    public DbTxt(String DbDir, String name)
    {
	this(DbDir, name, 0);
    }


/******************************************************************************
**
**     Purpose:                 Initializes the member variables to the
**                              parameters specified.
**    
**     Parameters:              String DbDir - database path
**                              String name - database name
**                              int contents - whats stored in db/txt file
**
**     Author/Last Modified:    Ram Periakaruppan, Sept 15th, 1999
**
******************************************************************************/

    public DbTxt(String DbDir, String name, int contents)
    {
	DbTxtName = name;
	this.contents = contents;
	if (DbTxtName.endsWith(".db"))
	{
	    type = DATABASE;

	    try
	    {
		gtraceDb = new GTraceDb(DbDir, DbTxtName);
	    }
	    catch(DbException dbe)
	    {
		System.out.println("Failed to open "
                                   + GTrace.DbDir+DbTxtName); 
	    }    
	}
	else 
	{
	    type = TEXTFILE;
	}
    }

//---------------------------------------------------------------------------
//                             public utility Methods
//---------------------------------------------------------------------------
    
/******************************************************************************
**
**     Method name:             lookup
**
**     Purpose:                 queries the database for the hostname/ipaddr
**    
**     Parameters:              String hostname - hostname
**                              String ipaddr - ip address
**
**     Return Values:           DbRecord - the result of the query
**                                           
**     Author/Last Modified:    Ram Periakaruppan, Sept. 15th, 1999
**
******************************************************************************/

    public DbRecord lookup(String hostname, String ipaddr) throws DbException
    {
	DbRecord record = null;
	
	if (contents == HOSTNAME || contents == HOSTNAMEIPADDR)
	{
	    record = gtraceDb.lookupkey(hostname);
	    if (record != null)
	    {
		record.setName(DbTxtName);
		return record;
	    }
	}
	
	if (contents == IPADDR || contents == HOSTNAMEIPADDR)
	{
	    record = gtraceDb.lookupkey(ipaddr);
	    if (record != null)
	    {
		record.setName(DbTxtName);
	    }
	}
	
	return record;
    }

/******************************************************************************
**
**     Method name:             lookup
**
**     Purpose:                 lookup up a key in the database and return
**                              the matching record
**    
**     Parameters:              String key - the key to be looked up
**
**     Return Values:           DbRecord - the matching record
**                                           
**     Author/Last Modified:    Ram Periakaruppan, Sept. 16th, 1999
**
******************************************************************************/

    public DbRecord lookup(String key) throws DbException
    {
	DbRecord record = null;
	
	record = gtraceDb.lookupkey(key);
	if (record != null)
	{
	    record.setName(DbTxtName);
	}
	return record;
    }

/******************************************************************************
**
**     Method name:             rlookup
**
**     Purpose:                 does a recursive lookup up a hostname, ie.
**                              trying to find a match for the possible
**                              combinations leading to the first level domain
**    
**     Parameters:              String hostname - the hostname to be looked up
**
**     Return Values:           DbRecord - the matching record
**                                           
**     Author/Last Modified:    Ram Periakaruppan, Sept. 15th, 1999
**
******************************************************************************/

    public DbRecord rlookup(String hostname) throws DbException
    {
	DbRecord record = null;
	
	String hstTokens[] = Util.tokenizeToString(hostname,".");

	if (hstTokens.length > 2)
	{
	    String hostnames[];
	    String lastDomain = Util.getDomain(hostname, 1);
	    int end=0;
	    if (lastDomain != null && lastDomain.length() == 2)
	    {
		hostnames = new String[hstTokens.length-3];
		end = hstTokens.length-3;
	    }
	    else
	    {
		hostnames = new String[hstTokens.length-2];
		end = hstTokens.length-2;
	    }
	    String lookupName = new String();

	    for(int i=1 ; i<=end ; i++)
	    {
		lookupName = "";
		for (int j=i; j<hstTokens.length; j++)
		{
		    if (j != hstTokens.length-1)
		    {
			lookupName = lookupName + hstTokens[j].toLowerCase()
                            + ".";
		    }
		    else
		    {
			lookupName = lookupName + hstTokens[j].toLowerCase();
		    }
		}
		hostnames[i-1] = lookupName;
	    }

	    for (int i=0; i<hostnames.length; i++)
	    {
		if (contents == HOSTNAME || contents == HOSTNAMEIPADDR)
		{
		    record = gtraceDb.lookupkey(hostnames[i]);
		    if (record != null)
		    {
			record.setName(DbTxtName);
			return record;
		    }
		}
	    }
	}
	
	return null;
    }

}
