/***************************************************************************
Copyright 1999 The Regents of the University of California
All Rights Reserved

Permission to use, copy, modify and distribute any part of the
GTrace tool for educational, research and non-profit
purposes, without fee, and without a written agreement is hereby
granted, provided that the above copyright notice, this paragraph
and the following paragraphs appear in all copies.

Those desiring to incorporate this into commercial products or use
for commercial purposes should contact the Technology Transfer
Office, University of California, San Diego, 9500 Gilman Drive, La
Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.

THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
ANY PATENT, TRADEMARK OR OTHER RIGHTS.

The GTrace tool is developed by Ram Periakaruppan at the University 
of California, San Diego under the Cooperative Association for Internet 
Data Analysis (CAIDA) Program. Support for this effort is provided by 
the CAIDA grant ANI-9711092, and by CAIDA members.

Report bugs and suggestions to gtrace-bugs@caida.org.
**************************************************************************/

/******************************************************************************
**
**     File:                    LocationDetails.java
**
**     Purpose:                 Used to store details about locations and their
**                              attributes.
**
**     Author/Date:             Ram Periakaruppan, Sept. 9th, 1999.
**
**     References:
**      Naughton, P., Schildt, H., JAVA - The Complete Reference. McGraw Hill.
**      Eckstein, R., Loy, M., Wood, D., JAVA Swing. O'Reilly.
**      The Source For Java Technology, http://www.java.sun.com.
**
*****************************************************************************/
package caida.GTrace;

import java.util.Hashtable;
import java.io.Serializable;

public class LocationDetails implements Serializable
{
    float lat;               // latitude
    float lon;               // longitude
    Integer searchStep;      // search step
    String formattedCity;    // formatted city
    String city;             // city
    String state;            // state
    String country;          // country
    String verifierResults;  // verifier results
    DbRecord dbRecord;       // database record
    RegexRecord regexRecord; // regex record  

    int hits; // the number of hits for the cache

    // Constants
    public static final int STEP1 = 10;
    public static final int STEP2 = 20;
    public static final int STEP3 = 30;
    public static final int STEP4 = 40;
    public static final int STEP4STATE = 41;
    public static final int STEP4COUNTRY = 42;
    public static final int STEP5 = 50;
    public static final int STEP6 = 60;
    public static final int STEP7 = 70;
    public static final int PRIVATE = 80;
    public static final int TIMEOUT = 90;

    final String STEP1STR = "Location obtained from cache";
    final String STEP2STR = "Location obatined from LOC record";
    final String STEP3STR = "There was a complete match for the "
                            + "hostname/ipaddress";
    final String STEP4STR = "Location obtained from rules in domain "
                            + "parsing file";
    final String STEP5STR = "Location obatined through a hostname "
                            + "reduction match";
    final String STEP6STR = "Location obtained from whois record via NetGeo";
    final String STEP7STR = "Mapped to Geographical center of country";
    final String PRIVATESTR = new String(
                                         "These following ranges are "
                                         + "reserved for private internet "
                                         + "use:\n\n" +
                                         "\t10.0.0.0 - 10.255.255.255\n" +
                                         "\t172.16.0.0.- 172.31.255.255\n" +
                                         "\t192.168.0.0 - 192.168.255.255\n");

//---------------------------------------------------------------------------
//                             constructors
//---------------------------------------------------------------------------
    
/******************************************************************************
 **
 **     Purpose:                 Initializes the member variables to their
 **                              default values.
 **    
 **     Parameters:              NONE
 **
 **     Author/Last Modified:    Ram Periakaruppan, Sept 9th, 1999
 **
 *****************************************************************************/

    public LocationDetails()
    {
	this.lat = 0;
	this.lon = 0;
	this.searchStep = new Integer(0);
	this.city = "";
	this.state = "";
	this.country = "";
        this.verifierResults = "";
	hits = 0;
    }

/******************************************************************************
 **
 **     Purpose:                 Initializes the member variables to the
 **                              parameters specified.
 **    
 **     Parameters:              Integer searchStep
 **
 **     Author/Last Modified:    Ram Periakaruppan, Sept 9th, 1999
 **
 *****************************************************************************/

    public LocationDetails(Integer searchStep)
    {
	this.lat = Float.POSITIVE_INFINITY;
	this.lon = Float.POSITIVE_INFINITY;
	this.searchStep = searchStep;
	this.city = "";
	this.state = "";
	this.country = "";
        this.verifierResults = "";
	hits = 0;
    }
    

/******************************************************************************
 **
 **     Purpose:                 Initializes the member variables to the
 **                              parameters specified.
 **    
 **     Parameters:              float lat - latitude
 **                              float lon - longitude
 **                              Integer searchStep - search step
 **                              String city - city
 **                              String state - state
 **                              String country - country
 **
 **     Author/Last Modified:    Ram Periakaruppan, Sept 9th, 1999
 **
 *****************************************************************************/

    public LocationDetails(float lat, float lon, Integer searchStep, 
			   String city, String state, String country)
    {
	this.lat = lat;
	this.lon = lon;
	this.searchStep = searchStep;
	this.city = city;
	this.state = state;
	this.country = country;
	hits = 0;

	if (city != null && state != null && country != null)
	{
	    formattedCity = new String(city + "," + state + "," + country);
	}
	else if (city != null && country != null)
  	{
	    formattedCity = new String(city + "," + country);
	}
	else if (country != null)
  	{
	    formattedCity = new String(country);
	}
	else
	{
	    formattedCity = new String("UNKNOWN");
	}
    }
    
/******************************************************************************
 **
 **     Purpose:                 Initializes the member variables to the
 **                              parameters specified.
 **    
 **     Parameters:              float lat - latitude
 **                              float lon - longitude
 **                              Integer searchStep - search step
 **
 **     Author/Last Modified:    Ram Periakaruppan, Sept 9th, 1999
 **
 *****************************************************************************/

    public LocationDetails(float lat, float lon, Integer searchStep)
    {
	this(lat,lon,searchStep,null,null,null);
	dbRecord = null;
	regexRecord = null;
    }

/******************************************************************************
 **
 **     Purpose:                 Initializes the member variables to the
 **                              parameters specified.
 **    
 **     Parameters:              float lat - latitude
 **                              float lon - longitude
 **                              Integer searchStep - search step
 **                              String data - the city,state and country
 **
 **     Author/Last Modified:    Ram Periakaruppan, Sept 9th, 1999
 **
 *****************************************************************************/

    public LocationDetails(float lat, float lon, Integer searchStep,
                           String data)
    {
	dbRecord = null;
	regexRecord = null;
	
	if (data != null)
	{
	    String tokens[] = Util.tokenizeToString(data,",");

 	    if (tokens.length == 3)
	    {
		setLocationDetails(lat,lon,searchStep,tokens[0]
                                   ,tokens[1],tokens[2]);
	    }
	    else if (tokens.length == 2)
	    {
		setLocationDetails(lat,lon,searchStep,tokens[0]
                                   ,null,tokens[1]);	
	    }
	    else if (tokens.length == 1)
	    {
		setLocationDetails(lat,lon,searchStep,null,null,tokens[0]); 
	    }
	}
	else
	{
	    setLocationDetails(lat,lon,searchStep,null,null,null);	
	}
    }


/******************************************************************************
 **
 **     Purpose:                 Initializes the member variables to the
 **                              parameters specified.
 **    
 **     Parameters:              DbRecord record - the database record
 **                              Integer searchStep - search step
 **
 **     Author/Last Modified:    Ram Periakaruppan, Sept 9th, 1999
 **
 *****************************************************************************/

    public LocationDetails(DbRecord record, Integer searchStep)
    {
	dbRecord = record;
	regexRecord = null;
	
	String dbFields[] = Util.tokenizeToString(dbRecord.getRecord(),"\t");

	if (dbFields.length > 2 && dbFields[2].length() > 0)
	{
	    String city = new String();
	    for (int i=2; i<dbFields.length; i++)
	    {
		city = city.concat(dbFields[i] + " ");
	    }

	    city = Util.formatCityLOC(city);

	    String tokens[] = Util.tokenizeToString(city,",");
	    
	    if (tokens.length == 3)
	    {
		setLocationDetails(Util.stringTofloat(dbFields[0]),
                                   Util.stringTofloat(dbFields[1]),
                                   searchStep,tokens[0],tokens[1],tokens[2]);
	    }
	    else if (tokens.length == 2)
	    {
		setLocationDetails(Util.stringTofloat(dbFields[0]),
                                   Util.stringTofloat(dbFields[1]),
                                   searchStep,tokens[0],null,tokens[1]);
	    }
	    else if (tokens.length == 1)
	    {
		setLocationDetails(Util.stringTofloat(dbFields[0]),
                                   Util.stringTofloat(dbFields[1]),
                                   searchStep,null,null,tokens[0]);	
	    }
	}
	else
	{
	    setLocationDetails(Util.stringTofloat(dbFields[0]),
			       Util.stringTofloat(dbFields[1]),
			       searchStep,null,null,null);	
	}
    }

/******************************************************************************
 **
 **     Purpose:                 Initializes the member variables to the
 **                              parameters specified.
 **    
 **     Parameters:              RegexRecord record - the regex record
 **
 **     Author/Last Modified:    Ram Periakaruppan, Sept 9th, 1999
 **
 *****************************************************************************/

    public LocationDetails(RegexRecord record)
    {
	dbRecord = null;
	regexRecord = record;
	DbRecord dbrecord;

	String dpfData = regexRecord.getDPFData();

	if (dpfData.startsWith("L:") || dpfData.startsWith("l:"))
	{
	    String dpfFields[] = Util.tokenizeToString(
                                                       dpfData.substring(
                                                  dpfData.indexOf(":")+1),",");
	    if (dpfFields.length == 2)
	    {
		setLocationDetails(Util.stringTofloat(dpfFields[0]),
                                   Util.stringTofloat(dpfFields[1]),
                                   new Integer(STEP4), null);
	    }
	}
	else if (dpfData.startsWith("country:"))
	{
	    LatLon country = GTrace.getCountryLatLon(
                                                     dpfData.substring(
                                                     dpfData.indexOf(":")+1));
	    if (country != null)
	    {
		setLocationDetails(country.getLat(),
                                   country.getLon(),
                                   new Integer(STEP4COUNTRY), null);
	    }
	}
	else
	{
	    dbrecord = regexRecord.getDbRecord();

	    if (dbrecord != null)
	    {
		String dbFields[] = Util.tokenizeToString(dbrecord.getRecord()
                                                          ,"\t");

		if (dbrecord.getName().compareTo(GTrace.AIRPORT) == 0)
		{
		    if (dbFields.length > 4 && dbFields[2].length() > 0)
		    {
                        String city = new String();
			for (int i=4; i<dbFields.length; i++)
			{
			    city = city.concat(dbFields[i] + " ");
			}
			
			city = Util.formatCityLOC(city);

			if (dbFields[2].compareTo("s") == 0)
			{
			    // STEP4STATE
			    setLocationDetails(Util.stringTofloat(dbFields[0]),
                                               Util.stringTofloat(dbFields[1]),
                                               new Integer(STEP4STATE), city);
			    return;
			}
			else if (dbFields[2].compareTo("c") == 0)
			{
			    // STEP4COUNTRY
			    setLocationDetails(Util.stringTofloat(dbFields[0]),
                                               Util.stringTofloat(dbFields[1]),
                                               new Integer(STEP4COUNTRY),city);
			    return;
			}
			else
			{
			    setLocationDetails(Util.stringTofloat(dbFields[0]),
                                               Util.stringTofloat(dbFields[1]),
                                               new Integer(STEP4), city);
			    return;
			}
		    }
		}
			    

		dpfData = Util.formatCityLOC(dbrecord.getKey());
		setLocationDetails(Util.stringTofloat(dbFields[0]),
                                   Util.stringTofloat(dbFields[1]),
                                   new Integer(STEP4), dpfData);
	    }
	}
    }


/******************************************************************************
 **
 **     Method name:             setLocationDetails
 **
 **     Purpose:                 set the member variables to the parameters
 **                              specified.
 **    
 **     Parameters:              float lat - latitude
 **                              float lon - longitude
 **                              Integer searchStep - search step
 **                              String city - city
 **                              String state - state
 **                              String country - country
 **
 **     Return Values:           void
 **                                           
 **     Author/Last Modified:    Ram Periakaruppan, Oct. 17th, 1999
 **
 *****************************************************************************/

    public void setLocationDetails(float lat, float lon, Integer searchStep, 
				   String city, String state, String country)
    {
	this.lat = lat;
	this.lon = lon;
	this.searchStep = searchStep;
	this.city = city;
	this.state = state;
	this.country = country;
	hits = 0;

	if (city != null && state != null && country != null)
	{
	    formattedCity = new String(city + "," + state + "," + country);
	}
	else if (city != null && country != null)
  	{
	    formattedCity = new String(city + "," + country);
	}
	else if (country != null)
  	{
	    formattedCity = new String(country);
	}
	else
	{
	    formattedCity = new String("UNKNOWN");
	}
    }

/******************************************************************************
 **
 **     Method name:             setLocationDetails
 **
 **     Purpose:                 set the member variables to the parameters
 **                              specified.
 **    
 **     Parameters:              float lat - latitude
 **                              float lon - longitude
 **                              Integer searchStep - search step
 **                              String data - the city,state and country
 **
 **     Return Values:           void
 **                                           
 **     Author/Last Modified:    Ram Periakaruppan, Oct. 17th, 1999
 **
 *****************************************************************************/

    public void setLocationDetails(float lat, float lon,
                                   Integer searchStep, String data)
    {
	if (data != null)
	{
	    String tokens[] = Util.tokenizeToString(data,",");

 	    if (tokens.length == 3)
	    {
		setLocationDetails(lat,lon,searchStep,tokens[0]
                                   ,tokens[1],tokens[2]);
	    }
	    else if (tokens.length == 2)
	    {
		setLocationDetails(lat,lon,searchStep,tokens[0]
                                   ,null,tokens[1]);	
	    }
	    else if (tokens.length == 1)
	    {
		setLocationDetails(lat,lon,searchStep,null,null,tokens[0]);
	    }
	}
	else
	{
	    setLocationDetails(lat,lon,searchStep,null,null,null);	
	}
    }

/******************************************************************************
 **
 **     Method name:             setVerifierResults
 **
 **     Purpose:                 set the verfier results to the parameter
 **                              specified.
 **    
 **     Parameters:              String data
 **
 **     Return Values:           void
 **                                           
 **     Author/Last Modified:    Ram Periakaruppan, Oct. 26th, 1999
 **
 *****************************************************************************/

    public void setVerifierResults(String data)
    {
        verifierResults = data;
    }
    
    
//---------------------------------------------------------------------------
//                             getxxx Methods
//---------------------------------------------------------------------------

/******************************************************************************
 **
 **     Method name:             getLat
 **
 **     Purpose:                 returns the latitude
 **    
 **     Parameters:              NONE
 **
 **     Return Values:           float - latitude
 **                                           
 **     Author/Last Modified:    Ram Periakaruppan, Oct. 17th, 1999
 **
 *****************************************************************************/

    public float getLat()
    {
	return lat;
    }

/******************************************************************************
 **
 **     Method name:             getLon
 **
 **     Purpose:                 returns the longitude
 **    
 **     Parameters:              NONE
 **
 **     Return Values:           float - longitude
 **                                           
 **     Author/Last Modified:    Ram Periakaruppan, Oct. 17th, 1999
 **
 *****************************************************************************/

    public float getLon()
    {
	return lon;
    }

/******************************************************************************
 **
 **     Method name:             getLatLonString
 **
 **     Purpose:                 returns a string representation of the latit-
 **                              ude and longitude
 **    
 **     Parameters:              NONE
 **
 **     Return Values:           String - string representation of lat/lon
 **                                           
 **     Author/Last Modified:    Ram Periakaruppan, Oct. 17th, 1999
 **
 *****************************************************************************/

    public String getLatLonString()
    {
	return Util.latlonToString(lat,lon);
    }

/******************************************************************************
 **
 **     Method name:             getCity
 **
 **     Purpose:                 returns the city
 **    
 **     Parameters:              NONE
 **
 **     Return Values:           String - city
 **                                           
 **     Author/Last Modified:    Ram Periakaruppan, Oct. 17th, 1999
 **
 *****************************************************************************/

    public String getCity()
    {
	return city;
    }

/******************************************************************************
 **
 **     Method name:             getState
 **
 **     Purpose:                 returns the state
 **    
 **     Parameters:              NONE
 **
 **     Return Values:           String - state
 **                                           
 **     Author/Last Modified:    Ram Periakaruppan, Oct. 17th, 1999
 **
 *****************************************************************************/

    public String getState()
    {
	return state;
    }

/******************************************************************************
 **
 **     Method name:             getCountry
 **
 **     Purpose:                 returns the country
 **    
 **     Parameters:              NONE
 **
 **     Return Values:           String - country
 **                                           
 **     Author/Last Modified:    Ram Periakaruppan, Oct. 17th, 1999
 **
 *****************************************************************************/

    public String getCountry()
    {
	return country;
    }

/******************************************************************************
 **
 **     Method name:             getSearchStep
 **
 **     Purpose:                 returns the search step that found the
 **                              location
 **    
 **     Parameters:              NONE
 **
 **     Return Values:           Integer - search step
 **                                           
 **     Author/Last Modified:    Ram Periakaruppan, Oct. 17th, 1999
 **
 *****************************************************************************/

    public Integer getSearchStep()
    {
	return searchStep;
    }

/******************************************************************************
 **
 **     Method name:             getHits
 **
 **     Purpose:                 returns the number of times that this location
 **                              detail object was accessed.
 **    
 **     Parameters:              NONE
 **
 **     Return Values:           int - the number of hits 
 **                                           
 **     Author/Last Modified:    Ram Periakaruppan, Oct. 17th, 1999
 **
 *****************************************************************************/

    public int getHits()
    {
	return hits;
    }

/******************************************************************************
 **
 **     Method name:             getFormattedCity
 **
 **     Purpose:                 returns a formatted city
 **    
 **     Parameters:              NONE
 **
 **     Return Values:           String - formatted representation of the city
 **                                           
 **     Author/Last Modified:    Ram Periakaruppan, Oct. 17th, 1999
 **
 *****************************************************************************/

    public String getFormattedCity()
    {
	return formattedCity;
    }


/******************************************************************************
 **
 **     Method name:             incrementHits
 **
 **     Purpose:                 increments the count of the number of times
 **                              this location was read from the cache.
 **    
 **     Parameters:              NONE
 **
 **     Return Values:           void
 **                                           
 **     Author/Last Modified:    Ram Periakaruppan, Oct. 17th, 1999
 **
 *****************************************************************************/
    
    public void incrementHits()
    {
        hits++;
    }
    
/******************************************************************************
 **
 **     Method name:             toString
 **
 **     Purpose:                 returns a string representation of the
 **                              contents of this object.
 **    
 **     Parameters:              NONE
 **
 **     Return Values:           String
 **                                           
 **     Author/Last Modified:    Ram Periakaruppan, Oct. 18th, 1999
 **
 *****************************************************************************/

    public String toString()
    {
	String retString;
	int sStep = searchStep.intValue();
	
	retString = new String(
			       "Latitude\t: " + lat + "\n" +
			       "Longitude\t: " + lon + "\n");

	if (city != null)
	{
	    retString = retString.concat("City\t: " + city + "\n");
	}

	if (state != null)
	{
	    retString = retString.concat("State\t: " + state + "\n");
	}

	if (country != null)
	{
	    retString = retString.concat("Country\t: " + country + "\n");
	}

	retString = retString.concat("\n");
	
	if (dbRecord != null)
	{
	    retString = retString.concat(dbRecord.toString());
	}

	if (regexRecord != null)
	{
	    retString = retString.concat(regexRecord.toString());
	}

        if (verifierResults != null && verifierResults.length() > 0)
        {
            retString = retString.concat("\nVerifier Results:\n");
            retString = retString.concat(verifierResults);
        }
        
	switch(sStep)
	{
        case STEP1:
            return STEP1STR + "\n\n" + retString;
        case STEP2:
            return STEP2STR + "\n\n" + retString;
        case STEP3:
            return STEP3STR + "\n\n" + retString;
        case STEP4:
        case STEP4STATE:
        case STEP4COUNTRY:
            return STEP4STR + "\n\n" + retString;
        case STEP5:
            return STEP5STR + "\n\n" + retString;
        case STEP6:
            return STEP6STR + "\n\n" + retString;
        case STEP7:
            return STEP7STR + "\n\n" + retString;
        case PRIVATE:
            return PRIVATESTR;
	}

	return retString;
    }
}
