/***************************************************************************
Copyright 1999 The Regents of the University of California
All Rights Reserved

Permission to use, copy, modify and distribute any part of the
GTrace tool for educational, research and non-profit
purposes, without fee, and without a written agreement is hereby
granted, provided that the above copyright notice, this paragraph
and the following paragraphs appear in all copies.

Those desiring to incorporate this into commercial products or use
for commercial purposes should contact the Technology Transfer
Office, University of California, San Diego, 9500 Gilman Drive, La
Jolla, CA 92093-0910, Ph: (619) 534-5815, FAX: (619) 534-7345.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
THE POSSIBILITY OF SUCH DAMAGE.

THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
ANY PATENT, TRADEMARK OR OTHER RIGHTS.

The GTrace tool is developed by Ram Periakaruppan at the University 
of California, San Diego under the Cooperative Association for Internet 
Data Analysis (CAIDA) Program. Support for this effort is provided by 
the CAIDA grant ANI-9711092, and by CAIDA members.

Report bugs and suggestions to gtrace-bugs@caida.org.
**************************************************************************/

/*****************************************************************************
* NetGeoClient.java
* 
* $Id: NetGeoClient.java,v 1.3 1999/10/27 00:05:36 ram Exp ram $
* 
* @author Jim Donohoe, CAIDA
* <LICENSE>
*****************************************************************************/
package caida.GTrace;

import java.net.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.Array;

public class NetGeoClient
{
   public static final String VERSION = "1.0";

   public static final String NETGEO_URL = 
      "http://netgeo.caida.org/perl/netgeo.cgi";
   
   // Status strings.
   public static final String OK = "OK";
   public static final String LOOKUP_IN_PROGRESS = "Lookup In Progress";
   public static final String NO_MATCH = "No Match";
   public static final String NO_COUNTRY = "No Country";
   public static final String UNKNOWN = "Unknown";

   public static final String WHOIS_ERROR = "WHOIS_ERROR";
   public static final String HTTP_ERROR = "HTTP_ERROR";

   // Strings for LAT_LONG_GRAN field.  The LAT_LONG_GRAN field will be
   // empty if the lat/long is (0,0). 
   public static final String CITY_GRAN = "City";
   public static final String STATE_GRAN = "State/Prov";
   public static final String COUNTRY_GRAN = "Country";

   // Indices into the 2-element latitude/longitude array.
   public static final int LAT_INDEX = 0;
   public static final int LONG_INDEX = 1;

   public static final boolean DEBUG = false;

   // String formed in the constructor from an application name supplied by
   // the invoker (optional), NetGeoClient name/version, the Java version
   // and vendor, and the operating system this code is running on.  This
   // string is URL encoded so it can be sent as a parameter in an HTTP GET 
   // query.  The NetGeo server will use the user-agent name to track any
   // platform-specific problems.  For example, the non-encoded string might
   // be: "plot_AS_loc/1.2.3 NetGeoClient/1.0 Sun Micr/1.1.7 Linux".
   private String _encodedUserAgentName;

//===========================================================================
//                             PUBLIC METHODS
//===========================================================================

/**
  * Description: 
  * 
  * @param String[] argv
  * @return void  
  * @author Jim Donohoe, CAIDA
  */
   public static void main( String[] argv )
   {
      NetGeoClient netgeo = new NetGeoClient( "TEST" );

      String target = "caida.org";
      System.out.println( "Testing getCountry( \"" + target + "\" )" );
      String country = netgeo.getCountry( target );
      System.out.println( "Result = " + country );
      System.out.println();
   }

//---------------------------------------------------------------------------
//                             constructors
//---------------------------------------------------------------------------

/**
  * Description: 
  * 
  * @param String - application name and version, e.g., "plot_AS_loc/1.2.3".
  * @return void  
  * @author Jim Donohoe, CAIDA
  */
   public NetGeoClient( String applicationName )
   {
      // Build up the user-agent name from the application name and version, 
      // the NetGeoClient name and version, the Java version and vendor, and
      // the operating system.  This name will be used to track any 
      // platform-specific problems. 
      StringBuffer buffer = new StringBuffer();
      if( applicationName != null && applicationName.length() > 0 )
      {
	 buffer.append( applicationName ).append( " " );
      }
      buffer.append( "NetGeoClient/" ).append( VERSION );
      String vendor = System.getProperty( "java.vendor" );

      // Some of the vendor strings can be really long so we truncate the 
      // string to a reasonable (and arbitrary) length.
      int maxVendorLength = 8;
      if( vendor != null && vendor.length() > maxVendorLength )
      {
         vendor = vendor.substring( 0, maxVendorLength );
      }
      buffer.append( " " ).append( vendor );
      buffer.append( "/" ).append(  System.getProperty( "java.version" ) );
      buffer.append( " " ).append( System.getProperty( "os.name" ) );

      // URL-encode the string and store it into this NetGeoClient object so 
      // it can be sent along with HTTP queries to the NetGeo server.  
      _encodedUserAgentName = URLEncoder.encode( buffer.toString() );
   }

/**
  * Description: 
  * 
  * @param 
  * @return void  
  * @author Jim Donohoe, CAIDA
  */
   public NetGeoClient()
   {
      // The user chose to not supply an application name and version,
      // so we invoke the regular constructor with an empty string.
      this( "" );
   }

//---------------------------------------------------------------------------
//                             scalar methods
//---------------------------------------------------------------------------

/**
  * Description: 
  * 
  * @param String
  * @return Hashtable
  * @author Jim Donohoe, CAIDA
  */
   public Hashtable getRecord( String target )
   {
      return execute( "getRecord", target );
   }

/**
  * Description: 
  * 
  * @param String
  * @return String
  * @author Jim Donohoe, CAIDA
  */
   public String getCountry( String target )
   {
      Hashtable partialRecord = execute( "getCountry", target );

      return (String) partialRecord.get( "COUNTRY" );
   }

/**
  * Description: Returns a hashtable containing ( String, String ) pairs,
  * with the keys LAT, LONG, LAT_LONG_GRAN, and STATUS.
  * 
  * @param String
  * @return Hashtable
  * @author Jim Donohoe, CAIDA
  */
   public Hashtable getLatLong( String target )
   {
      return execute( "getLatLong", target );
   }

//---------------------------------------------------------------------------
//                          getXxxArray methods
//---------------------------------------------------------------------------

/**
  * Description: 
  * 
  * @param String[]
  * @return Hashtable[]
  * @author Jim Donohoe, CAIDA
  */
   public Hashtable[] getRecordArray( String[] targetArray )
   {
      return execute( "getRecord", targetArray, false );
   }

/**
  * Description: 
  * 
  * @param String[]
  * @param boolean - nonblocking
  * @return Hashtable[]
  * @author Jim Donohoe, CAIDA
  */
   public Hashtable[] getRecordArray( String[] targetArray, 
				      boolean nonblocking )
   {
      return execute( "getRecord", targetArray, nonblocking );
   }

/**
  * Description: 
  * 
  * @param String[]
  * @return Hashtable[]
  * @author Jim Donohoe, CAIDA
  */
   public Hashtable[] getCountryArray( String[] targetArray )
   {
      return execute( "getCountry", targetArray, false );
   }

/**
  * Description: 
  * 
  * @param String[]
  * @param boolean - nonblocking
  * @return Hashtable[]
  * @author Jim Donohoe, CAIDA
  */
   public Hashtable[] getCountryArray( String[] targetArray, 
				       boolean nonblocking )
   {
      return execute( "getCountry", targetArray, nonblocking );
   }

/**
  * Description: 
  * 
  * @param String[]
  * @return Hashtable[]
  * @author Jim Donohoe, CAIDA
  */
   public Hashtable[] getLatLongArray( String[] targetArray )
   {
      return execute( "getLatLong", targetArray, false );
   }

/**
  * Description: 
  * 
  * @param String[]
  * @param boolean - nonblocking
  * @return Hashtable[]
  * @author Jim Donohoe, CAIDA
  */
   public Hashtable[] getLatLongArray( String[] targetArray, 
				       boolean nonblocking )
   {
      return execute( "getLatLong", targetArray, nonblocking );
   }

//---------------------------------------------------------------------------
//                         updateXxxArray methods
//---------------------------------------------------------------------------

/**
  * Description: 
  * 
  * @param Hashtable[]
  * @return String
  * @author Jim Donohoe, CAIDA
  */
   public String updateRecordArray( Hashtable[] recordArray )
   {
      return execute( "getRecord", recordArray, false );
   }

/**
  * Description: 
  * 
  * @param Hashtable[]
  * @param boolean - nonblocking
  * @return String
  * @author Jim Donohoe, CAIDA
  */
   public String updateRecordArray( Hashtable[] recordArray, 
				    boolean nonblocking )
   {
      return execute( "getRecord", recordArray, nonblocking );
   }

/**
  * Description: 
  * 
  * @param Hashtable[]
  * @return String
  * @author Jim Donohoe, CAIDA
  */
   public String updateCountryArray( Hashtable[] recordArray )
   {
      return execute( "getCountry", recordArray, false );
   }

/**
  * Description: 
  * 
  * @param Hashtable[]
  * @param boolean - nonblocking
  * @return String
  * @author Jim Donohoe, CAIDA
  */
   public String updateCountryArray( Hashtable[] recordArray, 
				     boolean nonblocking )
   {
      return execute( "getCountry", recordArray, nonblocking );
   }

/**
  * Description: 
  * 
  * @param Hashtable[]
  * @return String
  * @author Jim Donohoe, CAIDA
  */
   public String updateLatLongArray( Hashtable[] recordArray )
   {
      return execute( "getLatLong", recordArray, false );
   }

/**
  * Description: 
  * 
  * @param Hashtable[]
  * @param boolean - nonblocking
  * @return String
  * @author Jim Donohoe, CAIDA
  */
   public String updateLatLongArray( Hashtable[] recordArray, 
				     boolean nonblocking )
   {
      return execute( "getLatLong", recordArray, nonblocking );
   }

//---------------------------------------------------------------------------
//                    public static utility methods
//---------------------------------------------------------------------------

/**
  * Description: Convert an int array to an equivalent String array.  This
  * can be used to convert an array of AS numbers represented as ints to
  * the type required for input into getRecordArray, getCountryArray, etc.
  * 
  * @param int[]
  * @return String[]
  * @author Jim Donohoe, CAIDA
  */
   public static String[] intArrayToStringArray( int[] intArray )
   {
      String[] stringArray = null;
      if( intArray != null )
      {
	 stringArray = new String[ intArray.length ];

         for( int i = 0; i < intArray.length; i++ )
         {
	    stringArray[i] = Integer.toString( intArray[i] );
         }
      }
      return stringArray;
   }

/**
  * Description: Extract the latitude string from the lat/long hashtable
  * and return it as a float primitive.
  * 
  * @param Hashtable
  * @return float
  * @author Jim Donohoe, CAIDA
  */
   public static float getLat( Hashtable latLongHash )
   {
      return string2float( latLongHash, "LAT" );
   }


/**
  * Description: Extract the longitude string from the lat/long hashtable
  * and return it as a float primitive.
  * 
  * @param Hashtable
  * @return float
  * @author Jim Donohoe, CAIDA
  */
   public static float getLong( Hashtable latLongHash )
   {
      return string2float( latLongHash, "LONG" );
   }


/**
  * Description: Convert the Hashtable returned by getLatLong into a 
  * 2-element float array.
  * 
  * @param Hashtable - latLongHash
  * @return float[]
  * @author Jim Donohoe, CAIDA
  */
   public static float[] convertLatLong( Hashtable latLongHash )
   {
      float[] floatArray = new float[2];

      floatArray[LAT_INDEX] = string2float( latLongHash, "LAT" );
      floatArray[LONG_INDEX] = string2float( latLongHash, "LONG" );

      return floatArray;
   }


//===========================================================================
//                             PRIVATE METHODS
//===========================================================================

/**
  * Description: Method invoked from getXxx scalar methods.
  * 
  * @param String - methodName
  * @param String - target
  * @return Hashtable
  * @author Jim Donohoe, CAIDA
  */
   private Hashtable execute( String methodName, String target )
   {
      if( DEBUG )
      {
         System.out.println( methodName + "( '" + target + "' )" );
      }
      if( target == null || target.length() == 0 )
      {
	 System.err.println( "ERROR empty target for " + methodName );
      }

      String text = getTextFromNetGeoServer( methodName, target, false );

      Hashtable[] resultArray = convertToHashArray( text );

      Hashtable result = null;
      if( resultArray.length >= 1 )
      {
	 result = resultArray[0];
      }
      else
      {
	 // Something went wrong, perhaps the target string was empty.
	 // Return a hashtable with empty values so that getCountry and
	 // getLatLong can extract the empty values.  If the invoker is
	 // getRecord it will just return this hashtable it its invoker.
	 result = new Hashtable();
	 result.put( "COUNTRY", "" );
         result.put( "LAT", "0.0" );
	 result.put( "LONG", "0.0" );
	 result.put( HTTP_ERROR, "Unknown" );
      }
       
      return result;
   }


/**
  * Description: Method invoked from getXxxArray methods.
  * 
  * @param String - methodName
  * @param String[] - targetArray
  * @param boolean - nonblocking
  * @return Hashtable[]
  * @author Jim Donohoe, CAIDA
  */
   private Hashtable[] execute( String methodName, String[] targetArray, 
			        boolean nonblocking )
   {
      if( DEBUG )
      {
         System.out.println( makeInvokerString( "get", methodName,
						targetArray, nonblocking ) );
      }
      String targetList = makeList( methodName, targetArray );

      String text = getTextFromNetGeoServer( methodName, targetList, 
					     nonblocking );
      return convertToHashArray( text );
   }


/**
  * Description: Method invoked from updateXxxArray methods.
  * 
  * @param String - methodName
  * @param Hashtable[] - targetArray
  * @param boolean - nonblocking
  * @return String
  * @author Jim Donohoe, CAIDA
  */
   private String execute( String methodName, Hashtable[] inputArray, 
			   boolean nonblocking )
   {
      if( DEBUG )
      {
         System.out.println( makeInvokerString( "update", methodName,
						inputArray, nonblocking ) );
      }
      String targetList = makeList( methodName, inputArray );

      String text = getTextFromNetGeoServer( methodName, targetList, 
					     nonblocking );
      if( text.startsWith( HTTP_ERROR ) )
      {
	 // Return the full error message.
	 return text;
      }

      Hashtable[] resultArray = convertToHashArray( text );

      // Merge the new results into the existing record array.  The number
      // of entries in the two arrays may be different, but the relative
      // order of elements will be the same.
      // Example: Assume that records A, D, and E have already been found 
      // in an earlier lookup, so the identifer list would contain B, C, 
      // and F.  After the current lookup the result array will contain 
      // hashes for B, C, and F.
      // 
      //                  $inputIndex
      //                       |
      //                       V
      //                 +---+---+---+---+---+---+
      //  inputArray ->  | A | B | C | D | E | F |
      //                 +---+---+---+---+---+---+
      // 
      //              resultIndex
      //                   |
      //                   V
      //                 +---+---+---+
      //  resultArray -> | B | C | F |
      //                 +---+---+---+

      int inputIndex = 0;

      for( int resultIndex = 0; resultIndex < resultArray.length; 
	   resultIndex++ )
      {
	 // Extract the target string from the current result hash.
         String target = (String) resultArray[resultIndex].get( "TARGET" );

	 // Find the matching target in the input array, starting from the
	 // current input index.
	 while( inputIndex < inputArray.length &&
	        ! target.equals( inputArray[inputIndex].get( "TARGET" ) ) )
	 {
	    inputIndex += 1;
         }

	 if( inputIndex < inputArray.length )
	 {
	    // We matched an identifier in the input array, as expected.  
	    // Replace the hash in the input array with the hash returned 
	    // from the lookup.
	    inputArray[inputIndex] = resultArray[resultIndex];

	    // Advance to the next slot in the input array, no need to test 
	    // the current slot again.  
	    inputIndex += 1;
	 }
	 else
	 {
	    // Somehow we failed to find a matching target in the input array.
	    System.err.println ( "ERROR in updateArray, can't find '" +
				 target + "' in input array." );
	 }
      }

      return OK;
   }


/**
  * Description: 
  * 
  * @param String - methodName
  * @param String - targetString
  * @param boolean - nonblocking
  * @return String
  * @author Jim Donohoe, CAIDA
  */
   private String getTextFromNetGeoServer( String methodName, 
					   String targetString, 
					   boolean nonblocking )
   {
      if( targetString == null || targetString.length() == 0 )
      {
         // The target string is empty, no need to query the server. 
	 // Return an empty string, it will get converted into an empty hash
	 // by convertToHashArray.
	 if( DEBUG )
         {
	    System.out.println( "getTextFromNetGeoServer: Empty target" );
         }
	 return "";
      }

      // Form a URL with a GET query string.  The target string should be 
      // URL-encoded in case the user put in any embedded blanks or strange
      // characters.  For any correct target string the URL-encoded string
      // should be the same as the original, except that commas between 
      // targets get mapped to %2 in the encoded string, and a blank gets
      // mapped to +, as in "AS 1" -> "AS+1".
      String urlString = NETGEO_URL + "?method=" + methodName + 
	  "&target=" + URLEncoder.encode( targetString );
      if( nonblocking == true )
      {
	 urlString += "&nonblocking=true";
      }

      // Send the user-agent name as a query parameter.
      urlString += "&userAgent=" + _encodedUserAgentName;

      if( DEBUG )
      {
	 System.out.println( urlString );
      }

      URL url = null;
      try
      {
	 url = new URL( urlString );
      }
      catch( MalformedURLException e ) 
      {
	  System.err.println( "MalformedURLException: " + urlString );
	  e.printStackTrace();
      }

      String contentString = null;
      try
      {
	 // openConnection will return an object of HttpURLConnection when
	 // the URL is http.
	 HttpURLConnection netGeoConnection = 
	    (HttpURLConnection) url.openConnection();

	 // Attempt to connect to the NetGeo server URL
	 netGeoConnection.connect();

	 // Test the response code to see if the connection attempt was
	 // successful.
	 int responseCode = netGeoConnection.getResponseCode();
	 if( responseCode == HttpURLConnection.HTTP_OK )
	 {
	    // The connection attempt was successful.  Get an input stream
	    // from the connection and read the HTML text sent by the
	    // NetGeo server.
	    Object content = netGeoConnection.getContent();
	    if( content == null )
	    {
		System.err.println( "ERROR getContent returned null." );

	       // Set the content string to the empty string, not much more 
	       // we can do here.
	       contentString = "";
	    }
	    else if( InputStream.class.isAssignableFrom( content.getClass() ))
	    {
	       // The class of the content object is a subclass of InputStream,
	       // e.g., BufferedInputStream.
	       InputStream contentStream = (InputStream) content;

	       int bytesAvailable = contentStream.available();
	       byte[] byteArray = new byte[ bytesAvailable ];
 
	       // Read the bytes from the input stream and convert them into
	       // a string using the standard encoding, then store the string 
	       // into contentString for return to the invoker.
	       contentStream.read( byteArray, 0, bytesAvailable );
	       contentString = new String( byteArray, 0, bytesAvailable );
	    }
	    else if( content instanceof String )
	    {
	       contentString = (String) content;
	    }
	    else
	    {
	       // In some error situations the content object may be an object
	       // of type sun.net.www.content.text.PlainTextInputStream, which 
	       // is a hidden type.
	       System.err.println( "ERROR getContent returned an object " +
		  "of type " + content.getClass().getName() );

	       // Set the content string to the empty string, not much more 
	       // we can do here.
	       contentString = "";
	    }

	    // Close the connection to the NetGeo server.
	    netGeoConnection.disconnect();
         }
	 else
	 {
	    // The connection attempt failed.  Return the HTTP error code and
	    // error message, e.g., "HTTP_ERROR: 404 Not Found"
	    contentString = HTTP_ERROR + responseCode  + " " +
	       netGeoConnection.getResponseMessage();
	 }
      }  
      catch( IOException e ) 
      {
	  e.printStackTrace();
      }

      if( DEBUG )
      {
	 System.out.println( "Text from NetGeo server:\n" + contentString +
			     "\n----END OF TEXT----" );
      }

      if( contentString != null )
      {
	 // Trim the string.  This is only needed is the string is blank, so
	 // that the length==0 test in convertToHashArray works properly.
	 contentString = contentString.trim();
      }

      return contentString;
   }  


//---------------------------------------------------------------------------
//                    private static helper methods
//---------------------------------------------------------------------------

/**
  * Description: Helper method for getLat, getLong, convertLatLong. Extract 
  * the latitude string from the hashtable and return it as a float.
  * 
  * @param Hashtable - latLongHash
  * @param String - key
  * @return float
  * @author Jim Donohoe, CAIDA
  */
   private static float string2float( Hashtable latLongHash, String key )
   {
      float floatValue = (float) 0.0;
      if( latLongHash != null )
      {  
	 String stringValue = (String) latLongHash.get( key );
	 try
	 {
	    floatValue = new Float( stringValue ).floatValue();
	 }
         catch( NumberFormatException e )
	 {
	    System.err.println( "ERROR in string2float, " + 
	       "NumberFormatException: string = '" + stringValue + "'" );
	    e.printStackTrace();
	 }
      }
      else
      {
	 System.err.println( "ERROR in string2float, null hashtable input" );
      }
      return floatValue;
   }


/**
  * Description: 
  * 
  * @param String[]
  * @return String
  * @author Jim Donohoe, CAIDA
  */
   private static String makeList(  String methodName, String[] stringArray )
   {
      StringBuffer list = new StringBuffer();

      if( stringArray == null || stringArray.length == 0 )
      {
	 // methodName is the name of the scalar method.  Print a message
	 // such as "ERROR empty argument array for getRecordArray".
	 System.err.println( "ERROR empty target array for " + 
			     methodName + "Array" );
      }
      else
      {	 
         // The input target array is not empty.  Loop through the array,
	 // adding target strings to the string buffer.
	 for( int i = 0; i < stringArray.length; i++ )
	 {
	    String target = stringArray[i];
	    if( target != null )
	    {
	       target = target.trim();
	       if( ! target.equals( stringArray[i] ) )
	       {
	          System.err.println( "WARNING trimming target string '" +
				      stringArray[i] + "' at index = " + i );
		  // Store the trimmed string into the array, so the strings
		  // in the input array will match the results.
		  stringArray[i] = target;
	       }
	       if( target.length() > 0 )
	       {
		  if( i > 0 )
		  {
		     list.append( "," );
		  }
		  list.append( target );
	       }
	    }
	    if( target == null || target.length() == 0 )
	    {
	       System.err.println( "ERROR empty target at index = " + i +
		                   " for " + methodName + "Array" );
	    }
	 }
      }
      return list.toString();
   }


/**
  * Description: 
  * 
  * @param Hashtable[]
  * @return String
  * @author Jim Donohoe, CAIDA
  */
   private static String makeList(  String methodName, Hashtable[] hashArray )
   {
      StringBuffer list = new StringBuffer();

      if( hashArray == null || hashArray.length == 0 )
      {
	 // methodName is the name of the scalar method.  Convert it to the
	 // name of an updateXxxArray method to report the error, e.g.,
	 // "ERROR empty argument array for updateRecordArray".
	 System.err.println( "ERROR empty target array for " + 
			     "update" + methodName.substring( 3 ) + "Array" );
      }
      else
      {	 
	 // The input target array is not empty.  Loop through the array,
	 // adding target strings to the string buffer.
	 for( int i = 0; i < hashArray.length; i++ )
	 {
	    if( hashArray[i] != null )
	    {
	       // Check the status to see if this record has already been
	       // successfully looked up.  The user may be polling the server
	       // asking for updates of the records which haven't been found
	       // yet.
	       String status = (String) hashArray[i].get( "STATUS" );

	       // status will be null when it hasn't been set yet, e.g., as
	       // when an updateXxxArray method gets called for the first time
	       // with hashes containing only target strings.  status will be
	       // LOOKUP_IN_PROGRESS if a previous lookup failed to find the
	       // target in the database and the whois lookup is proceeding.
	       // If the status string is something else it means the target
	       // has already been found or the whois lookups failed, so there
	       // is no need to look it up again.
	       if( status == null || status.equals( LOOKUP_IN_PROGRESS ) )
	       {
	          String target = (String) hashArray[i].get( "TARGET" );
	          if( target != null )
	          {
		     // Use a temp var trimmedTarget so we can keep a ref
		     // to the original and not need to do another hash
		     // lookup to compare to the trimmed value.
	             String trimmedTarget = target.trim();
		     if( ! trimmedTarget.equals( target ) )
		     {
	                System.err.println( "WARNING trimming target " +
			   "string '" + target + "' at index = " + i );
			// Store the trimmed string into the hash, so the 
			// strings in the input array will match the results.
			// This will overwrite the existing target string.
		        hashArray[i].put( "TARGET", trimmedTarget );
	             }
		     target = trimmedTarget;
	             if( target.length() > 0 )
	             {
		        if( i > 0 )
		        {
		           list.append( "," );
		        }
		        list.append( target );
	             }
	          }
	          if( target == null || target.length() == 0 )
	          {
	             System.err.println( "ERROR empty target at index = " + i 
		     + " for update" + methodName.substring( 3 ) + "Array" );
	          }
	       }
	    }
	    else
	    {
	       System.err.println( "ERROR empty hashtable at index = " + i +
		    " for update" + methodName.substring( 3 ) + "Array" );
	    }
	 }
      }
      return list.toString();
   }


/**
  * Description: Parse the block of text output from netgeo.cgi, such as:
  * <!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML//EN">
  * <HTML><HEAD><TITLE>NetGeo Results</TITLE>
  * </HEAD><BODY>
  * TARGET:        caida.org<br>
  * COUNTRY:       US<br>
  * STATUS:        OK<br>
  * </BODY></HTML>
  * 
  * The input text might be an HTTP error message, indicated by HTTP_ERROR.
  * If the text starts with HTTP_ERROR this method will return a 1-element
  * array containing a hashtable with key HTTP_ERROR and using the input 
  * text (error code and message) as the value.
  * 
  * @param String
  * @return Hashtable[]
  * @author Jim Donohoe, CAIDA
  */
   private static Hashtable[] convertToHashArray( String text )
   {
      // First, check for an HTTP error message or an empty string.
      if( text.startsWith( HTTP_ERROR ) || text.length() == 0 )
      {
         // An error occurred in the HTTP connection.  Make a 1-element array
         // for returning the error message to the original invoker.
         Hashtable[] errorMessageArray = new Hashtable[1];
         errorMessageArray[0] = new Hashtable();

	 if( text.length() == 0 )
	 {
	    text = "Empty content string";
	 }

         // Use the HTTP_ERROR constant as the key so it can be easily tested.
         errorMessageArray[0].put( HTTP_ERROR, text );

         return errorMessageArray;
      }

      // Process the non-error text, storing each block of key/value pairs 
      // into a hashtable and storing the hashtables into a vector.

      Vector hashVector = new Vector();
      Hashtable hashtable = null;

      StringTokenizer tokenizer = new StringTokenizer( text, "\n" );
      while( tokenizer.hasMoreTokens() )
      {
	 String line = tokenizer.nextToken();

	 // All key/value lines will be like: "TARGET:        caida.org<br>"
	 // with a key in all uppercase followed by :, then at least one
	 // space, then the value and the <br>.
	 String key = null;
	 String value = null;
	 int colonIndex = line.indexOf( ':' );
	 if( colonIndex > 0 )
	 {
	    // Get the key string from the start of the line.
	    key = line.substring( 0, colonIndex );

	    // Find the index of the HTML tag at the end of the line.
	    int tagIndex = line.indexOf( "<br>" );
	    if( tagIndex > 0 )
	    {
	       value = line.substring( colonIndex+1, tagIndex ).trim();
	    }
	 }
	 
         // Check to see if we found a key/value pair.  The value string might
	 // be the empty string but shouldn't be null if key is not null.
	 if( key != null && value != null )
	 {
	    if( key.equals( "TARGET" ) )
	    {
	       // This is the start of a new block, start a new hashtable.
	       hashtable = new Hashtable();
	       hashtable.put( "TARGET", value );
	    }
	    else if( key.equals( "STATUS" ) )
	    {
               // This is the end of a block, add the STATUS key/value pair
	       // then store the current hashtable into the vector.
	       if( hashtable != null )
	       {
		  hashtable.put( "STATUS", value );
		  hashVector.addElement( hashtable );
		  hashtable = null;
	       }
	    }
	    else
	    {
	       // Add the key/value pair to the current hashtable.  
	       if( hashtable != null )
	       {
		  hashtable.put( key, value );
	       }
	    }
	 }
      }
      
      // Convert the vector of hashtables into an array of hashtables.
      Hashtable[] hashArray = new Hashtable[ hashVector.size() ];
      hashVector.copyInto( hashArray );

      return hashArray;
   }


/**
  * Description: Method used when DEBUG == true to print invoker of 
  * execute method and print short description of arg list.
  * 
  * @param String - type, "get" or "update"
  * @param String - methodName
  * @param Hashtable[]
  * @return String
  * @author Jim Donohoe, CAIDA 
  */
   private static String makeInvokerString(  String type, String methodName, 
					   Object array, boolean nonblocking )
   {
      String invokerString = null;
      String argList = "( null, " + nonblocking + " )";
      if( array != null )
      {
	 argList = "( [" + Array.getLength( array ) + " elements], " + 
	     nonblocking + " )";
      }
	   
      if( type.equals( "get" ) )
      {
	 invokerString = methodName + "Array" + argList;
      }
      else
      {
	 invokerString = "update" + methodName.substring( 3 ) + "Array" +
	    argList;
      }
      return invokerString;
   }
}
